__all__ = ['prepare_coords', 'report']
from typing import Optional, Sequence, List
from copy import deepcopy
from ipywidgets import widgets as wdg
from pandas import DataFrame
from ._distance import get_topics_dist, get_topics_scatter
from ._vis import plot_scatter_topics, plot_terms, plot_docs
from ._helpers import (
    calc_terms_probs_ratio,
    get_phi, get_theta,
    get_top_docs)


def prepare_coords(
        model: object,
        labels: Optional[Sequence] = None,
        dist_kws: dict = None,
        scatter_kws: dict = None) -> DataFrame:
    """Prepare coordinates for topics scatter plot.

    Parameters
    ----------
    model : object
        Topic model instance.
    labels : Optional[Sequence]
        Topics labels.
    dist_kws : dict, optional
        Keyword arguments passed to :py:meth:`tmplot.get_topics_dist()`.
    scatter_kws : dict, optional
        Keyword arguments passed to :py:meth:`tmplot.get_topics_scatter()`.
    """
    if not dist_kws:
        dist_kws = {}
    if not scatter_kws:
        scatter_kws = {}

    phi = get_phi(model)
    theta = get_theta(model)
    topics_dists = get_topics_dist(phi, **dist_kws)
    topics_coords = get_topics_scatter(topics_dists, theta, **scatter_kws)
    topics_coords['label'] = labels or theta.index
    return topics_coords


def report(
        model: object = None,
        docs: Optional[Sequence[str]] = None,
        topics_labels: Optional[Sequence[str]] = None,
        corpus: Optional[List] = None,
        layout: wdg.Layout = None,
        show_headers: bool = True,
        show_docs: bool = True,
        show_words: bool = True,
        show_topics: bool = True,
        topics_kws: dict = None,
        height: int = 500,
        width: int = 250,
        coords_kws: dict = None,
        words_kws: dict = None,
        docs_kws: dict = None,
        top_docs_kws: dict = None) -> wdg.VBox:
    """Interactive report interface.

    Parameters
    ----------
    model : object, optional
        Topic model instance.
    docs : Optional[Sequence[str]], optional
        Documents.
    topics_labels : Optional[Sequence[str]], optional
        Topics labels.
    corpus : Optional[List[str]], optional
        Gensim corpus.
    layout : wdg.Layout, optional
        Interface layout instance.
    show_headers : bool, optional
        Show headers.
    show_docs : bool, optional
        Show documents widget.
    show_words : bool, optional
        Show words widget.
    show_topics : bool, optional
        Show topics scatter plot widget.
    topics_kws : dict, optional
        Keyword arguments passed to :py:meth:`tmplot.plot_scatter_topics()`.
    coords_kws : dict, optional
        Keyword arguments passed to :py:meth:`tmplot.prepare_coords()`.
    words_kws : dict, optional
        Keyword arguments passed to :py:meth:`tmplot.plot_terms()`.
    docs_kws : dict, optional
        Keyword arguments passed to :py:meth:`tmplot.plot_docs()`.
    top_docs_kws : dict, optional
        Keyword arguments passed to :py:meth:`tmplot.get_top_docs()`.

    Returns
    -------
    wdg.VBox
        Report interface as a VBox instance.
    """
    from IPython.display import display

    _topics_kws = {
        'chart_kws': {'height': height, 'width': width}}\
        if not topics_kws else deepcopy(topics_kws)
    _coords_kws = {} if not coords_kws else deepcopy(coords_kws)
    _words_kws = {
        'chart_kws': {'height': height, 'width': width}}\
        if not words_kws else deepcopy(words_kws)
    _top_docs_kws = {} if not docs_kws else deepcopy(top_docs_kws)
    _docs_kws = {} if not docs_kws else deepcopy(docs_kws)

    # Headers init
    topics_header = wdg.HTML('<b>Topics scatter plot</b>')\
        if show_headers and show_topics else None
    words_header = wdg.HTML('<b>Relevant words (terms)</b>')\
        if show_headers and show_words else None
    docs_header = wdg.HTML('<b>Top documents in a topic</b>')\
        if show_headers and show_docs else None

    # Layout init
    grid_cols = " ".join(['1fr'] * sum([show_docs, show_words, show_topics]))
    layout = wdg.Layout(
            grid_template_columns=grid_cols,
            # justify_items='center'
        )\
        if not layout else layout

    # Children widgets list init
    children = []

    if 'topics_coords' not in _topics_kws:
        topics_coords = prepare_coords(model, **_coords_kws)
        _topics_kws.update({
            'topics_coords': topics_coords,
            'label_col': 'label',
            'size_col': 'size',
            'topic': 0
        })

    if 'terms_probs' not in _words_kws:
        phi = get_phi(model)
        terms_probs = calc_terms_probs_ratio(phi, topic=0)
        _words_kws.update({'terms_probs': terms_probs})

    if 'docs' not in _docs_kws:
        theta = get_theta(model, corpus=corpus).values
        _top_docs_kws.update({
            'docs': docs, 'theta': theta,
            'topics': [0], 'docs_num': 2})
        top_docs = get_top_docs(**_top_docs_kws)
        top_docs.columns = ['']
        _docs_kws.update({'docs': top_docs})

    # Topic selection
    def _on_select_topic(sel):
        topic = sel['new']
        topics_plot_output.clear_output(wait=False)
        words_plot_output.clear_output(wait=False)
        docs_plot_output.clear_output(wait=False)
        with words_plot_output:
            terms_probs = calc_terms_probs_ratio(
                phi, topic=topic, lambda_=lambda_slider.value)
            _words_kws.update({'terms_probs': terms_probs})
            display(plot_terms(**_words_kws))
        with topics_plot_output:
            _topics_kws.update({'topic': topic})
            display(plot_scatter_topics(**_topics_kws))
        with docs_plot_output:
            _top_docs_kws.update({'topics': [sel['new']]})
            top_docs = get_top_docs(**_top_docs_kws)
            top_docs.columns = ['']
            _docs_kws.update({'docs': top_docs})
            display(plot_docs(**_docs_kws))

    topics_ids = list(range(len(_topics_kws['topics_coords'])))
    topics_labels = topics_labels or topics_ids
    select_topic = wdg.Dropdown(
        options=list(zip(topics_labels, topics_ids)), value=0)
    select_topic.observe(_on_select_topic, names='value')
    select_topic_header = wdg.HTML('<b>Select a topic</b>:')
    select_topic_widget = wdg.HBox([select_topic_header, select_topic])
    select_topic_wrapper = wdg.VBox(
        [select_topic_widget], layout={'align_items': 'center'})

    # Topics scatter
    def _on_select_topics_method(names):
        topics_plot_output.clear_output(wait=False)
        with topics_plot_output:
            _coords_kws.update({'scatter_kws': {'method': names['new']}})
            topics_coords = prepare_coords(model, **_coords_kws)
            _topics_kws.update({
                'topics_coords': topics_coords,
                'topic': select_topic.value
            })
            display(plot_scatter_topics(**_topics_kws))

    if show_topics:
        topics_plot_children = [topics_header] if show_headers else []
        options_methods = [
            ('TSNE', 'tsne'),
            ('SpectralEmbedding', 'sem'),
            ('MDS', 'mds'),
            ('LocallyLinearEmbedding (Standard)', 'lle'),
            ('LocallyLinearEmbedding (LTSA)', 'ltsa'),
            ('Isomap', 'isomap')
        ]
        topics_method_header = wdg.HTML('Select a method:')
        topics_method = wdg.Dropdown(
            options=options_methods,
            value='tsne',
            layout=wdg.Layout(width=f'{width/1.25}px')
        )
        topics_method_widget = wdg.HBox([topics_method_header, topics_method])
        topics_method.observe(_on_select_topics_method, names='value')
        topics_plot_output = wdg.Output()
        topics_plot = plot_scatter_topics(**_topics_kws)
        topics_plot_output.append_display_data(topics_plot)
        topics_plot_children.extend([topics_method_widget, topics_plot_output])
        topics_widget = wdg.VBox(
            topics_plot_children,
            layout={'align_items': 'center'})
        children.append(topics_widget)

    # Words
    if show_words:
        def _on_select_lambda(sel):
            topic = select_topic.value
            lambda_ = lambda_slider.value
            words_plot_output.clear_output(wait=False)
            with words_plot_output:
                terms_probs = calc_terms_probs_ratio(
                    phi, topic=topic, lambda_=lambda_)
                _words_kws.update({'terms_probs': terms_probs})
                display(plot_terms(**_words_kws))

        lambda_slider = wdg.FloatSlider(
            value=0.6,
            min=0.0,
            max=1.0,
            step=0.01,
            description='',
            continuous_update=False,
            orientation='horizontal',
            readout=True,
            readout_format='.2f',
            layout=wdg.Layout(width=f'{width/1.25}px')
        )
        lambda_slider.observe(_on_select_lambda, names='value')
        lambda_slider_header = wdg.HTML('Lambda value:')
        lambda_slider_widget = wdg.HBox([lambda_slider_header, lambda_slider])
        words_plot = plot_terms(**_words_kws)
        words_plot_output = wdg.Output()
        words_plot_output.append_display_data(words_plot)
        words_plot_children = [words_header, lambda_slider_widget]\
            if show_headers else [lambda_slider_widget]
        words_plot_children.append(words_plot_output)
        words_widget = wdg.VBox(
            words_plot_children,
            layout={'align_items': 'center'})
        children.append(words_widget)

    # Docs
    if show_docs:
        def _on_select_docs_num(_):
            docs_num = docs_num_slider.value
            docs_plot_output.clear_output(wait=False)
            with docs_plot_output:
                _top_docs_kws.update({'docs_num': docs_num})
                top_docs = get_top_docs(**_top_docs_kws)
                top_docs.columns = ['']
                _docs_kws.update({'docs': top_docs})
                display(plot_docs(**_docs_kws))

        docs_num_slider = wdg.IntSlider(
            value=2,
            min=1,
            max=100,
            continuous_update=False,
            orientation='horizontal',
            readout=True,
            readout_format='d',
            layout=wdg.Layout(width=f'{width/1.25}px')
        )
        docs_num_slider.observe(_on_select_docs_num, names='value')
        docs_num_slider_header = wdg.HTML('Documents number:')
        docs_num_slider_widget = wdg.HBox(
            [docs_num_slider_header, docs_num_slider])

        docs_plot = plot_docs(**_docs_kws)
        docs_plot_output = wdg.Output()
        docs_plot_output.append_display_data(docs_plot)
        docs_plot_children = [docs_header, docs_num_slider_widget]\
            if show_headers else [docs_num_slider_widget]
        docs_plot_children.append(docs_plot_output)
        docs_widget = wdg.VBox(
            docs_plot_children,
            layout={'align_items': 'center'})
        children.append(docs_widget)

    grid_box = wdg.GridBox(children, layout=layout)
    hr_line = wdg.HTML('<hr style="border: 0; border-bottom: 1px solid #aaa">')
    app = wdg.VBox([select_topic_wrapper, hr_line, grid_box])

    return app



__all__ = [
    'get_topics_dist', 'get_topics_scatter', 'get_top_topic_words']
from typing import Union, List
from itertools import combinations
from pandas import DataFrame
import numpy as np
from scipy.special import kl_div
from scipy.spatial import distance
from sklearn.manifold import (
    TSNE, Isomap, LocallyLinearEmbedding, MDS, SpectralEmbedding)
from ._helpers import calc_topics_marg_probs


def _dist_klb(a1: np.ndarray, a2: np.ndarray):
    dist = kl_div(a1, a2)
    return dist[np.isfinite(dist)].sum()


def _dist_sklb(a1: np.ndarray, a2: np.ndarray):
    dist = kl_div(a1, a2) + kl_div(a1, a2)
    return dist[np.isfinite(dist)].sum()


def _dist_jsd(a1: np.ndarray, a2: np.ndarray):
    dist = 0.5 * kl_div(a1, a2) + 0.5 * kl_div(a1, a2)
    return dist[np.isfinite(dist)].sum()


def _dist_jef(a1: np.ndarray, a2: np.ndarray):
    vals = (a1 - a2) * (np.log(a1) - np.log(a2))
    vals[(vals <= 0) | ~np.isfinite(vals)] = 0.
    return vals.sum()


def _dist_hel(a1: np.ndarray, a2: np.ndarray):
    a1[(a1 <= 0) | ~np.isfinite(a1)] = 1e-64
    a2[(a2 <= 0) | ~np.isfinite(a2)] = 1e-64
    hel_val = distance.euclidean(
        np.sqrt(a1), np.sqrt(a2)) / np.sqrt(2)
    return hel_val


def _dist_bhat(a1: np.ndarray, a2: np.ndarray):
    pq = a1 * a2
    pq[(pq <= 0) | ~np.isfinite(pq)] = 1e-64
    dist = -np.log(np.sum(np.sqrt(pq)))
    return dist


def _dist_tv(a1: np.ndarray, a2: np.ndarray):
    dist = np.sum(np.abs(a1 - a2)) / 2
    return dist


def _dist_jac(a1: np.ndarray, a2: np.ndarray,  top_words=100):
    a = np.argsort(a1)[:-top_words-1:-1]
    b = np.argsort(a2)[:-top_words-1:-1]
    j_num = np.intersect1d(a, b, assume_unique=False).size
    j_den = np.union1d(a, b).size
    jac_val = 1 - j_num / j_den
    return jac_val


def get_topics_dist(
        phi: Union[np.ndarray, DataFrame],
        method: str = "sklb",
        **kwargs) -> np.ndarray:
    """Finding closest topics in models.

    Parameters
    ----------
    phi : Union[ndarray, DataFrame]
        Words vs topics matrix (W x T).
    method : str = "sklb"
        Comparison method. Possible variants:
        1) "klb" - Kullback-Leibler divergence.
        2) "sklb" - Symmetric Kullback-Leibler divergence.
        3) "jsd" - Jensen-Shannon divergence.
        4) "jef" - Jeffrey's divergence.
        5) "hel" - Hellinger distance.
        6) "bhat" - Bhattacharyya distance.
        7) "tv" — Total variation distance.
        8) "jac" - Jaccard index.
    **kwargs : dict
        Keyword arguments passed to distance function.

    Returns
    -------
    numpy.ndarray
        Topics distances matrix.
    """
    phi_copy = np.array(phi)
    topics_num = phi_copy.shape[1]
    topics_pairs = combinations(range(topics_num), 2)

    # Topics distances matrix
    topics_dists = np.zeros(shape=(topics_num, topics_num), dtype=float)

    dist_funcs = {
        "klb": _dist_klb,
        "sklb": _dist_sklb,
        "jsd": _dist_jsd,
        "jef": _dist_jef,
        "hel": _dist_hel,
        "bhat": _dist_bhat,
        "tv": _dist_tv,
        "jac": _dist_jac,
    }

    for i, j in topics_pairs:
        _dist_func = dist_funcs.get(method, "sklb")
        topics_dists[((i, j), (j, i))] = _dist_func(
            phi_copy[:, i], phi_copy[:, j], **kwargs)

    return topics_dists


def get_topics_scatter(
        topic_dists: np.ndarray,
        theta: np.ndarray,
        method: str = 'tsne',
        method_kws: dict = None) -> DataFrame:
    """Calculate topics coordinates for a scatter plot.

    Parameters
    ----------
    topic_dists : numpy.ndarray
        Topics distance matrix.
    theta : numpy.ndarray
        Topics vs documents probability matrix.
    method : str = 'tsne'
        Method to calculate topics scatter coordinates (X and Y).
        Possible values:
        1) 'tsne' - t-distributed Stochastic Neighbor Embedding.
        2) 'sem' - SpectralEmbedding.
        3) 'mds' - MDS.
        4) 'lle' - LocallyLinearEmbedding.
        5) 'ltsa' - LocallyLinearEmbedding with LTDA method.
        6) 'isomap' - Isomap.
    method_kws : dict = None
        Keyword arguments passed to method function.

    Returns
    -------
    DataFrame
        Topics scatter coordinates.
    """
    if not method_kws:
        method_kws = {'n_components': 2}

    if method == 'tsne':
        method_kws.setdefault('metric',  'precomputed')
        transformer = TSNE(**method_kws)

    elif method == 'sem':
        method_kws.setdefault('affinity', 'precomputed')
        transformer = SpectralEmbedding(**method_kws)

    elif method == 'mds':
        method_kws.setdefault('dissimilarity', 'precomputed')
        transformer = MDS(**method_kws)

    elif method == 'lle':
        method_kws['method'] = 'standard'
        transformer = LocallyLinearEmbedding(**method_kws)

    elif method == 'ltsa':
        method_kws['method'] = 'ltsa'
        transformer = LocallyLinearEmbedding(**method_kws)

    elif method == 'isomap':
        transformer = Isomap(**method_kws)

    coords = transformer.fit_transform(topic_dists)

    topics_xy = DataFrame(coords, columns=['x', 'y'])
    topics_xy['topic'] = topics_xy.index.astype(int)
    topics_xy['size'] = calc_topics_marg_probs(theta)
    size_sum = topics_xy['size'].sum()
    if size_sum > 0:
        topics_xy['size'] *= (100 / topics_xy['size'].sum())
    else:
        topics_xy['size'] = np.nan
    return topics_xy


def get_top_topic_words(
        phi: DataFrame,
        words_num: int = 20,
        topics_idx: Union[List[int], np.ndarray] = None) -> DataFrame:
    """Select top topic words from a fitted model.

    Parameters
    ----------
    phi : pandas.DataFrame
        Words vs topics matrix (phi) with words as
        indices and topics as columns.
    words_num : int = 20
        The number of words to select.
    topics_idx : Union[List, numpy.ndarray] = None
        Topics indices.

    Returns
    -------
    DataFrame
        Words with highest probabilities in all (or selected) topics.
    """
    return phi.loc[:, topics_idx or phi.columns]\
        .apply(
            lambda x: x
            .sort_values(ascending=False)
            .head(words_num).index, axis=0
    )



from ._helpers import *  # noqa: F401, F403
from ._distance import *  # noqa: F401, F403
from ._report import *  # noqa: F401, F403
from ._stability import *  # noqa: F401, F403
from ._vis import *  # noqa: F401, F403

__version__ = '0.0.7'



# TODO: heatmap of docs in topics
# TODO: topic dynamics in time
# TODO: word cloud
__all__ = [
    'plot_scatter_topics', 'plot_terms', 'plot_docs']
from typing import Union, Sequence
from pandas import DataFrame, option_context
from numpy import ndarray
from altair import (
    AxisConfig, Chart, X, Y, Size, Color, value, Text, Scale, Legend)


def plot_scatter_topics(
        topics_coords: Union[ndarray, DataFrame],
        x_col: str = "x",
        y_col: str = "y",
        topic: int = None,
        size_col: str = None,
        label_col: str = None,
        color_col: str = None,
        topic_col: str = None,
        font_size: int = 13,
        x_kws: dict = None,
        y_kws: dict = None,
        chart_kws: dict = None,
        circle_kws: dict = None,
        circle_enc_kws: dict = None,
        text_kws: dict = None,
        text_enc_kws: dict = None,
        size_kws: dict = None,
        color_kws: dict = None) -> Chart:
    """Topics scatter plot in 2D.

    Parameters
    ----------
    topics_coords : Union[ndarray, DataFrame]
        Topics scatter coordinates.
    x_col : str, optional
        X column name.
    y_col : str, optional
        Y column name.
    topic : int, optional
        Topic index.
    size_col : str, optional
        Column with size values.
    label_col : str, optional
        Column with topic labels.
    color_col : str, optional
        Column with colors.
    topic_col : str, optional
        Column with topics texts.
    font_size : int, optional
        Font size.
    x_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.X()`.
    y_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Y()`.
    chart_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Chart()`.
    circle_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Chart.mark_circle()`.
    circle_enc_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Chart.encode()`
        for circle elements.
    text_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Chart.mark_text()`.
    text_enc_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Chart.encode()`
        for text elements.
    size_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Size()` for
        circle elements.
    color_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Color()` for
        circle elements.

    Returns
    -------
    Chart
        Topics scatter plot.
    """
    if not chart_kws:
        chart_kws = {}

    if not x_kws:
        x_kws = {'shorthand': x_col}

    if not y_kws:
        y_kws = {'shorthand': y_col}

    if not circle_kws:
        circle_kws = {"opacity": 0.33, "stroke": 'black', "strokeWidth": 1}

    if not size_kws:
        size_kws = {
            'title': 'Marginal topic distribution',
            'scale': Scale(range=[0, 3000])}

    if not circle_enc_kws:
        circle_enc_kws = {
            "x": X(**x_kws),
            "y": Y(**y_kws),
            "size": Size(size_col, **size_kws)
            if size_col and not topics_coords[size_col].isna().any()
            else value(500)
        }

    if not text_kws:
        text_kws = {"align": "center", "baseline": "middle"}

    if not color_kws:
        color_kws = {}\
            if topic is None\
            else {'condition': {
                "test": f"datum['topic'] == {topic}", "value": "red"}}

    data = DataFrame(topics_coords, columns=[x_col, y_col])\
        if isinstance(topics_coords, ndarray)\
        else topics_coords.copy()

    if not topic_col:
        topic_col = "topic"
        data = data.assign(**{topic_col: range(len(topics_coords))})

    if not text_enc_kws:
        text_enc_kws = {
            "x": X(**x_kws),
            "y": Y(**y_kws),
            "text": Text(topic_col),
            "size": value(font_size)}

    # Tooltips initialization
    tooltips = []
    if label_col:
        tooltips.append(label_col)
    if size_col:
        tooltips.append(size_col)

    if tooltips:
        circle_enc_kws.update({'tooltip': tooltips})
        text_enc_kws.update({'tooltip': tooltips})

    if color_kws:
        circle_enc_kws.update({'color': Color(**color_kws)})

    base = Chart(data, **chart_kws)

    rule = base\
        .mark_rule()\
        .encode(
            y='average(y)',
            color=value('gray'),
            size=value(0.2))

    rule2 = base\
        .mark_rule()\
        .encode(
            x='average(x)',
            color=value('gray'),
            size=value(0.2))

    points = base\
        .mark_circle(**circle_kws)\
        .encode(**circle_enc_kws)

    text = base\
        .mark_text(**text_kws)\
        .encode(**text_enc_kws)

    return (rule + rule2 + points + text)\
        .configure_axis(labelFontSize=font_size, titleFontSize=font_size, grid=False)\
        .configure(axis=AxisConfig(disable=True))\
        .configure_view(stroke='transparent', strokeWidth=0)\
        .configure_legend(
            orient='bottom',
            labelFontSize=font_size,
            titleFontSize=font_size)


def plot_terms(
        terms_probs: DataFrame,
        x_col: str = 'Probability',
        y_col: str = 'Terms',
        color_col: str = 'Type',
        font_size: int = 13,
        chart_kws: dict = None,
        bar_kws: dict = None,
        x_kws: dict = None,
        y_kws: dict = None,
        color_kws: dict = None) -> Chart:
    """Plot words conditional and marginal probabilities.

    Parameters
    ----------
    terms_probs : DataFrame
        Words probabilities.
    x_col : str, optional
        X column name.
    y_col : str, optional
        Y column name.
    color_col : str, optional
        Column with values types (for coloring).
    font_size : int, optional
        Font size.
    chart_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Chart()`.
    bar_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Chart.mark_bar()`.
    x_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.X()`.
    y_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Y()`.
    color_kws : dict, optional
        Keyword arguments passed to :py:meth:`altair.Color()`.

    Returns
    -------
    Chart
        Terms probabilities chart.
    """
    if not x_kws:
        x_kws = {'stack': None}
    if not y_kws:
        y_kws = {'sort': None, 'title': None}
    if not color_kws:
        color_kws = {
            'shorthand': color_col,
            'legend': Legend(orient='bottom'),
            'scale': Scale(scheme='category20')
        }
    if not chart_kws:
        chart_kws = {}
    if not bar_kws:
        bar_kws = {}

    return Chart(data=terms_probs, **chart_kws)\
        .mark_bar(**bar_kws)\
        .encode(
            x=X(x_col, **x_kws),
            y=Y(y_col, **y_kws),
            color=Color(**color_kws)
    )\
        .configure_axis(labelFontSize=font_size, titleFontSize=font_size)\
        .configure_legend(
            labelFontSize=font_size, titleFontSize=font_size,
            columns=1, labelLimit=250)


def plot_docs(
        docs: Union[Sequence[str], DataFrame],
        styles: str = None,
        html_kws: dict = None) -> DataFrame:
    """Documents plotting functionality for report interface.

    Parameters
    ----------
    docs : Union[Sequence[str], DataFrame]
        Documents.
    styles : str, optional
        Styles string for formatting the table with documents.
        Concatenated with HTML.
    html_kws : dict, optional
        Keyword arguments passed to :py:meth:`pandas.DataFrame.to_html` method.

    Returns
    -------
    ipywidgets.HTML
        Topic documents.
    """
    from IPython.display import HTML

    if styles is None:
        styles = '<style>table td{text-align: left !important}' +\
            'table th{text-align: center !important}</style>'
    if html_kws is None:
        # html_kws = {'classes': 'plot'}
        html_kws = {}

    if isinstance(docs, DataFrame):
        df_docs = docs.copy()
    else:
        df_docs = DataFrame({'docs': docs})

    with option_context('display.max_colwidth', 0):
        # df_docs.style.set_properties(**{'text-align': 'center'})
        return HTML(styles + df_docs.to_html(**html_kws))



__all__ = [
    'get_phi', 'get_theta',
    'get_relevant_terms', 'get_salient_terms',
    'get_docs', 'get_top_docs',
    'calc_terms_marg_probs', 'calc_topics_marg_probs',
    'calc_terms_probs_ratio']
from warnings import warn
from importlib.util import find_spec
from typing import Union, Optional, Sequence, List
from functools import partial
from math import log
from numpy import ndarray, zeros, argsort, array, arange, vstack
from pandas import concat, Series, DataFrame

tomotopy_installed = find_spec('tomotopy')
if tomotopy_installed:
    from tomotopy import (
        LDAModel as tomotopyLDA,
        LLDAModel as tomotopyLLDA,
        CTModel as tomotopyCT,
        DMRModel as tomotopyDMR,
        HDPModel as tomotopyHDP,
        PTModel as tomotopyPT,
        SLDAModel as tomotopySLDA,
        GDMRModel as tomotopyGDMR)

gensim_installed = find_spec('gensim')
if gensim_installed:
    from gensim.models.ldamodel import LdaModel as gensimLDA
    from gensim.models.ldamulticore import LdaMulticore as gensimLDAMC

bitermplus_installed = find_spec('bitermplus')
if bitermplus_installed:
    from bitermplus._btm import BTM


def __warn_package_installation(package_name: str):
    warn(
        f'Please install "{package_name}" package to analyze its models.\n'
        f'Run `pip install {package_name}` in the console.')


def get_phi(
        model: object,
        vocabulary: Optional[Sequence] = None) -> DataFrame:
    """Get words vs topics matrix (phi).

    Returns ``phi`` matrix of shape W x T, where W is the number of words,
    and T is the number of topics.

    Parameters
    ----------
    model : object
        Topic model instance.
    vocabulary : Optional[Sequence], optional
        Vocabulary as a list of words. Needed for getting ``phi`` matrix
        from ``gensim`` model instance.

    Returns
    -------
    DataFrame
        Words vs topics matrix (phi).
    """
    phi = None

    if _is_tomotopy(model):

        # Topics vs words distributions
        twd = list(map(model.get_topic_word_dist, range(model.k)))

        # Concatenating into DataFrame
        phi = DataFrame(vstack(twd).T)

        # Specifying terms from vocabulary as index
        phi.index = model.vocabs

    elif _is_gensim(model):

        phi = DataFrame(model.get_topics().T)
        if vocabulary:
            phi.index = vocabulary

    elif _is_btmplus(model):
        phi = model.df_words_topics_

    if isinstance(phi, DataFrame):
        phi.index.name = 'words'
        phi.columns.name = 'topics'

    return phi


def _is_tomotopy(model: object) -> bool:
    if tomotopy_installed:
        tomotopy_models = [
            tomotopyLDA, tomotopyLLDA, tomotopyCT, tomotopyDMR, tomotopyHDP,
            tomotopyPT, tomotopySLDA, tomotopyGDMR]
        return any(map(partial(isinstance, model), tomotopy_models))
    else:
        __warn_package_installation("tomotopy")
        return False


def _is_gensim(model: object) -> bool:
    if gensim_installed:
        gensim_models = [gensimLDA, gensimLDAMC]
        return any(map(partial(isinstance, model), gensim_models))
    else:
        __warn_package_installation("gensim")
        return False


def _is_btmplus(model: object) -> bool:
    if bitermplus_installed:
        return isinstance(model, BTM)
    else:
        __warn_package_installation("bitermplus")
        return False


def get_theta(
        model: object,
        corpus: Optional[List] = None) -> DataFrame:
    """Get topics vs documents (theta) matrix.

    Returns theta matrix of shape T x D, where T is the number of topics,
    D is the number of documents.

    Parameters
    ----------
    model : object
        Topic model instance.
    corpus : Optional[List], optional
        Corpus.

    Returns
    -------
    DataFrame
        Topics vs documents matrix (theta).
    """
    theta = None

    if _is_tomotopy(model):
        tdd = list(map(lambda x: x.get_topic_dist(), model.docs))
        theta = DataFrame(vstack(tdd).T)

    elif _is_gensim(model):
        if corpus is None:
            raise ValueError(
                '`corpus` must be supplied for a gensim model')
        tdd = list(map(model.get_document_topics, corpus))
        theta = DataFrame(zeros((len(tdd), model.num_topics)))
        for doc_id, doc_topic in enumerate(tdd):
            for topic_id, topic_prob in doc_topic:
                theta.loc[doc_id, topic_id] = topic_prob
        theta = theta.T

    elif _is_btmplus(model):
        theta = DataFrame(model.matrix_topics_docs_)

    if isinstance(theta, DataFrame):
        theta.index.name = 'topics'
        theta.columns.name = 'docs'

    return theta


def get_docs(
        model: object) -> List[str]:
    """Retrieve documents from topic model object.

    Parameters
    ----------
    model : object
        Topic model instance.

    Returns
    -------
    List[str]
        List of documents.
    """
    if _is_tomotopy(model):
        docs_raw = map(lambda x: x.words, model.docs)
        return list(
            map(
                lambda doc: " ".join(map(lambda x: model.vocabs[x], doc)),
                docs_raw))
    return None


def get_top_docs(
        docs: Sequence[str],
        model: object = None,
        theta: ndarray = None,
        corpus: Optional[List] = None,
        docs_num: int = 5,
        topics: Sequence[int] = None) -> DataFrame:
    """Get top documents for all (or a selected) topic.

    Parameters
    ----------
    docs : Sequence
        List of documents.
    model : object, optional
        Topic model instance.
    theta : ndarray, optional
        Topics vs documents matrix.
    corpus : Optional[List], optional
        Corpus for ``gensim`` model.
    docs_num : int, optional
        Number of documents to return.
    topics : Sequence[int], optional
        Sequence of topics indices.

    Returns
    -------
    DataFrame
        Top documents.

    Raises
    ------
    ValueError
        If neither a model or theta matrix is passed, ValueError is raised.
    """
    if all([model is None, theta is None]):
        raise ValueError("Please pass a model or a theta matrix to function")

    if model and not theta:
        theta = get_theta(model, corpus=corpus).values

    def _select_docs(docs, theta, topic_id: int):
        probs = theta[topic_id, :]
        idx = argsort(probs)[:-docs_num-1:-1]
        result = Series(list(map(lambda x: docs[x], idx)))
        result.name = 'topic{}'.format(topic_id)
        return result

    topics_num = theta.shape[0]
    topics_idx = arange(topics_num) if topics is None else topics
    return concat(
        map(lambda x: _select_docs(docs, theta, x), topics_idx), axis=1)


def calc_topics_marg_probs(
        theta: Union[DataFrame, ndarray],
        topic_id: int = None) -> Union[DataFrame, ndarray]:
    """Calculate marginal topics probabilities.

    Parameters
    ----------
    theta : Union[DataFrame, ndarray]
        Topics vs documents matrix.
    topic_id : int, optional
        Topic index.

    Returns
    -------
    Union[DataFrame, ndarray]
        Marginal topics probabilities.
    """
    if topic_id:
        if isinstance(theta, ndarray):
            return theta[topic_id, :].sum()
        elif isinstance(theta, DataFrame):
            return theta.iloc[topic_id, :].sum()
    else:
        return theta.sum(axis=1)


def calc_terms_marg_probs(
        phi: Union[ndarray, DataFrame],
        word_id: Optional[int] = None) -> Union[ndarray, Series]:
    """Calculate marginal terms probabilities.

    Parameters
    ----------
    phi : Union[ndarray, DataFrame]
        Words vs topics matrix.
    word_id: Optional[int]
        Word index.

    Returns
    -------
    Union[ndarray, Series]
        Marginal terms probabilities.
    """
    if word_id:
        if isinstance(phi, ndarray):
            return phi[word_id, :].sum()
        elif isinstance(phi, DataFrame):
            return phi.iloc[word_id, :].sum()
    else:
        return phi.sum(axis=1)


def get_salient_terms(
        terms_freqs: ndarray,
        phi: ndarray,
        theta: ndarray) -> ndarray:
    """Get salient terms.

    Calculated as:
    saliency(w) = frequency(w) * [sum_t p(t | w) * log(p(t | w)/p(t))],
    where ``w`` is a term index, ``t`` is a topic index.

    Parameters
    ----------
    terms_freqs : numpy.ndarray
        Words frequencies.
    phi : numpy.ndarray
        Words vs topics matrix.
    theta : numpy.ndarray
        Topics vs documents matrix.

    Returns
    -------
    numpy.ndarray
        Terms saliency values.
    """
    p_t = array(calc_topics_marg_probs(theta))
    p_w = array(calc_terms_marg_probs(phi))

    def _p_tw(phi, w, t):
        return phi[w, t] * p_t[t] / p_w[w]

    saliency = array([
        terms_freqs[w] * sum([
            _p_tw(phi, w, t) * log(_p_tw(phi, w, t) / p_t[t])
            for t in range(phi.shape[1])])
        for w in range(phi.shape[0])
    ])
    # saliency(term w) = frequency(w)
    # * [sum_t p(t | w) * log(p(t | w)/p(t))] for topics t
    # p(t | w) = p(w | t) * p(t) / p(w)
    return saliency


def calc_terms_probs_ratio(
        phi: DataFrame,
        topic: int,
        terms_num: int = 30,
        lambda_: float = 0.6) -> DataFrame:
    """Get terms conditional and marginal probabilities.

    Parameters
    ----------
    phi : DataFrame
        Words vs topics matrix.
    topic : int
        Topic index.
    terms_num : int, optional
        Number of words to return.
    lambda_ : float, optional
        Weight parameter. It determines the weight given to the probability
        of term W under topic T relative to its lift [1]_. Setting it to 1
        equals topic-specific probabilities of terms.

    References
    ----------
    .. [1] Sievert, C., & Shirley, K. (2014). LDAvis: A method for visualizing
           and interpreting topics. In Proceedings of the workshop on
           interactive language learning, visualization, and interfaces (pp.
           63-70).

    Returns
    -------
    DataFrame
        Words conditional and marginal probabilities.
    """
    p_cond_name = 'Conditional term probability, p(w|t)'
    p_cond = phi.iloc[:, topic]\
        .rename(p_cond_name)\
        if isinstance(phi, DataFrame)\
        else Series(phi[:, topic], name=p_cond_name)

    p_marg_name = 'Marginal term probability, p(w)'
    p_marg = phi.sum(axis=1)\
        .rename(p_marg_name)\
        if isinstance(phi, DataFrame)\
        else Series(phi[:, topic], name=p_marg_name)

    terms_probs = concat((p_marg, p_cond), axis=1)
    relevant_idx = get_relevant_terms(phi, topic, lambda_).index
    terms_probs_slice = terms_probs.loc[relevant_idx].head(terms_num)

    return terms_probs_slice\
        .reset_index(drop=False)\
        .melt(
            id_vars=[terms_probs_slice.index.name],
            var_name='Type',
            value_name='Probability')\
        .rename(columns={terms_probs_slice.index.name: 'Terms'})


def get_relevant_terms(
        phi: Union[ndarray, DataFrame],
        topic: int,
        lambda_: float = 0.6) -> Series:
    """Select relevant terms.

    Parameters
    ----------
    phi : Union[np.ndarray, DataFrame]
        Words vs topics matrix (phi).
    topic : int
        Topic index.
    lambda_ : float = 0.6
        Weight parameter. It determines the weight given to the probability
        of term W under topic T relative to its lift [2]_. Setting it to 1
        equals topic-specific probabilities of terms.

    References
    ----------
    .. [2] Sievert, C., & Shirley, K. (2014). LDAvis: A method for visualizing
           and interpreting topics. In Proceedings of the workshop on
           interactive language learning, visualization, and interfaces (pp.
           63-70).

    Returns
    -------
    Series
        Terms sorted by relevance (descendingly).
    """
    phi_topic = phi.iloc[:, topic]\
        if isinstance(phi, DataFrame)\
        else phi[:, topic]
    # relevance = lambda * p(w | t) + (1 - lambda) * p(w | t)/p(w)
    relevance = lambda_ * phi_topic\
        + (1 - lambda_) * phi_topic / phi.sum(axis=1)
    return relevance.sort_values(ascending=False)



__all__ = ['get_closest_topics', 'get_stable_topics']
from typing import List, Tuple, Any
import numpy as np
import tqdm
from ._distance import _dist_klb, _dist_sklb, _dist_jsd, _dist_jef, _dist_hel,\
    _dist_bhat, _dist_jac, _dist_tv
from ._helpers import get_phi

dist_funcs = {
    "klb": _dist_klb,
    "sklb": _dist_sklb,
    "jsd": _dist_jsd,
    "jef": _dist_jef,
    "hel": _dist_hel,
    "bhat": _dist_bhat,
    "tv": _dist_tv,
    "jac": _dist_jac,
}


def get_closest_topics(
        models: List[Any],
        ref: int = 0,
        method: str = "sklb",
        top_words: int = 100,
        verbose: bool = True) -> Tuple[np.ndarray, np.ndarray]:
    """Finding closest topics in models.

    Parameters
    ----------
    models : List[Any]
        List of supported and fitted topic models.
    ref : int = 0
        Index of reference matrix (zero-based indexing).
    method : str = "sklb"
        Distance calculation method. Possible variants:
        1) "klb" - Kullback-Leibler divergence.
        2) "sklb" - Symmetric Kullback-Leibler divergence.
        3) "jsd" - Jensen-Shannon divergence.
        4) "jef" - Jeffrey's divergence.
        5) "hel" - Hellinger distance.
        6) "bhat" - Bhattacharyya distance.
        7) "tv" - Total variation distance.
        8) "jac" - Jaccard index.
    top_words : int = 100
        Number of top words in each topic to use in Jaccard index calculation.
    verbose : bool = True
        Verbose output (progress bar).

    Returns
    -------
    closest_topics : np.ndarray
        Closest topics indices in one two-dimensional array.
        Columns correspond to the compared models (their indices),
        rows are the closest topics pairs.
    dist : np.ndarray
        Kullback-Leibler (if ``method`` is set to ``klb``) or Jaccard index
        values corresponding to the matrix of the closest topics.

    Example
    -------
    >>> # `models` must be an iterable of fitted models
    >>> closest_topics, kldiv = tmplot.get_closest_topics(models)
    """
    # Number of models passed
    models_num = len(models)

    # Reference model id
    ref = models_num - 1 if ref >= models_num else ref

    # Reference model
    model_ref = models[ref]

    # Words vs topics matrix (phi)
    model_ref_phi = get_phi(model_ref)

    # Number of topics
    print(type(model_ref_phi))
    topics_num = model_ref_phi.shape[1]

    # Array with the closest topics ids
    closest_topics = np.zeros(shape=(topics_num, models_num), dtype=int)
    closest_topics[:, ref] = np.arange(topics_num)

    # Distance function selection
    dist_func = dist_funcs.get(method, "sklb")

    # Distance values
    dist_vals = np.zeros(shape=(topics_num, models_num), dtype=float)

    def enum_func(x):
        return enumerate(tqdm.tqdm(x)) if verbose else enumerate(x)

    # Iterating over all models
    for mid, model in enum_func(models):

        # Current model is equal to reference model, skipping
        if mid == ref:
            continue

        # Distance matrix for all topic pairs
        all_vs_all_dists = np.zeros((topics_num, topics_num))

        # Iterating over all topic pairs
        for t_ref in range(topics_num):
            for t in range(topics_num):
                all_vs_all_dists[t_ref, t] = dist_func(
                    model_ref_phi.iloc[t_ref, :], get_phi(model).iloc[t, :])

        # Creating two arrays for the closest topics ids and distance values
        if method == "jac":
            closest_topics[:, mid] = np.argmax(all_vs_all_dists, axis=1)
            dist_vals[:, mid] = np.max(all_vs_all_dists, axis=1)
        else:
            closest_topics[:, mid] = np.argmin(all_vs_all_dists, axis=1)
            dist_vals[:, mid] = np.min(all_vs_all_dists, axis=1)

    return closest_topics, dist_vals


def get_stable_topics(
        closest_topics: np.ndarray,
        dist: np.ndarray,
        norm: bool = True,
        inverse: bool = True,
        inverse_factor: float = 1.0,
        ref: int = 0,
        thres: float = 0.9,
        thres_models: int = 2) -> Tuple[np.ndarray, np.ndarray]:
    """Finding stable topics in models.

    Parameters
    ----------
    closest_topics : np.ndarray
        Closest topics indices in a two-dimensional array.
        Columns correspond to the compared matrices (their indices),
        rows are the closest topics pairs. Typically, this should be
        the first value returned by :meth:`tmplot.get_closest_topics`
        function.
    dist : np.ndarray
        Distance values: Kullback-Leibler divergence or Jaccard index values
        corresponding to the matrix of the closest topics.
        Typically, this should be the second value returned by
        :meth:`tmplot.get_closest_topics` function.
    norm : bool = True
        Normalize distance values (passed as ``dist`` argument).
    inverse : bool = True
        Inverse distance values by subtracting them from ``inverse_factor``.
        Should be set to ``False`` if Jaccard index was used to calculate
        closest topics.
    inverse_factor : float = 1.0
        Subtract distance values from this factor to inverse.
    ref : int = 0
        Index of reference matrix (i.e. reference column index,
        zero-based indexing).
    thres : float = 0.9
        Threshold for distance values filtering.
    thres_models : int = 2
        Minimum topic recurrence frequency across all models.

    Returns
    -------
    stable_topics : np.ndarray
        Filtered matrix of the closest topics indices (i.e. stable topics).
    dist : np.ndarray
        Filtered distance values corresponding to the matrix of
        the closest topics.

    See Also
    --------
    tmplot.get_closest_topics

    Example
    -------
    >>> closest_topics, kldiv = tmplot.get_closest_topics(models)
    >>> stable_topics, stable_kldiv = tmplot.get_stable_topics(
    ...     closest_topics, kldiv)
    """
    dist_arr = np.asarray(dist)
    dist_ready = dist_arr / dist_arr.max() if norm else dist_arr.copy()
    dist_ready = inverse_factor - dist_ready if inverse else dist_ready
    mask = (
        np.sum(np.delete(dist_ready, ref, axis=1) >= thres, axis=1)
        >= thres_models)
    return closest_topics[mask], dist_ready[mask]
